from typing import Sequence

from libcst import Arg, Call, Integer

from django_codemod.constants import DJANGO_3_1, DJANGO_4_0
from django_codemod.utils.calls import find_keyword_arg
from django_codemod.visitors.base import BaseFuncRenameTransformer


class GetRandomStringTransformer(BaseFuncRenameTransformer):
    """Update lengthless uses of get_random_string()"""

    deprecated_in = DJANGO_3_1
    removed_in = DJANGO_4_0
    rename_from = "django.utils.crypto.get_random_string"
    rename_to = "django.utils.crypto.get_random_string"

    # The default value Django specifies is 12.
    default_length_value = Integer("12")

    def update_call_args(self, node: Call) -> Sequence[Arg]:
        # No args? Just add the default 12.
        if not node.args:
            return [Arg(value=self.default_length_value)]

        # If there's only an allowed chars kwarg, prepend the length arg.
        allowed_chars_kwarg = find_keyword_arg(node.args, "allowed_chars")

        if len(node.args) == 1 and allowed_chars_kwarg:
            return [
                Arg(value=self.default_length_value),
                allowed_chars_kwarg,
            ]

        # Otherwise don't do anything.
        return node.args



from typing import Sequence

from libcst import Arg, Call, Name
from libcst.codemod.visitors import AddImportsVisitor

from django_codemod.constants import DJANGO_2_2, DJANGO_3_1
from django_codemod.visitors.base import BaseFuncRenameTransformer


class FixedOffsetTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.timezone.FixedOffset` by `datetime.timezone`."""

    deprecated_in = DJANGO_2_2
    removed_in = DJANGO_3_1
    rename_from = "django.utils.timezone.FixedOffset"
    rename_to = "datetime.timezone"

    def update_call_args(self, node: Call) -> Sequence[Arg]:
        """Update first argument to convert integer for minutes to timedelta."""
        AddImportsVisitor.add_needed_import(
            context=self.context,
            module="datetime",
            obj="timedelta",
        )
        offset_arg, *other_args = node.args
        timedelta_call = Call(
            func=Name("timedelta"),
            args=(Arg(keyword=Name("minutes"), value=offset_arg.value),),
        )
        new_offset_arg = offset_arg.with_changes(value=timedelta_call)
        return (new_offset_arg, *other_args)



from django_codemod.constants import DJANGO_2_2, DJANGO_3_1, DJANGO_4_0
from django_codemod.visitors.base import BaseFuncRenameTransformer


class FloatRangeModelFieldTransformer(BaseFuncRenameTransformer):
    """Replace postgres' model field `FloatRangeField` by `DecimalRangeField`."""

    deprecated_in = DJANGO_2_2
    removed_in = DJANGO_3_1
    rename_from = "django.contrib.postgres.fields.FloatRangeField"
    rename_to = "django.contrib.postgres.fields.DecimalRangeField"


class FloatRangeFormFieldTransformer(BaseFuncRenameTransformer):
    """Replace postgres' form field `FloatRangeField` by `DecimalRangeField`."""

    deprecated_in = DJANGO_2_2
    removed_in = DJANGO_3_1
    rename_from = "django.contrib.postgres.forms.FloatRangeField"
    rename_to = "django.contrib.postgres.forms.DecimalRangeField"


class JSONModelFieldTransformer(BaseFuncRenameTransformer):
    """Replace postgres' JSON field with the core JSON field."""

    deprecated_in = DJANGO_3_1
    removed_in = DJANGO_4_0
    rename_from = "django.contrib.postgres.fields.JSONField"
    rename_to = "django.db.models.JSONField"



from django_codemod.constants import DJANGO_2_2, DJANGO_3_1
from django_codemod.visitors.base import BaseFuncRenameTransformer


class QuerySetPaginatorTransformer(BaseFuncRenameTransformer):
    """Replace `QuerySetPaginator` class by `Paginator`."""

    deprecated_in = DJANGO_2_2
    removed_in = DJANGO_3_1
    rename_from = "django.core.paginator.QuerySetPaginator"
    rename_to = "django.core.paginator.Paginator"



"""Module to implement base functionality."""
from abc import ABC
from typing import List, Optional, Sequence, Tuple, Union

from libcst import (
    Arg,
    Attribute,
    BaseExpression,
    BaseSmallStatement,
    Call,
    CSTNode,
    ImportAlias,
    ImportFrom,
    ImportStar,
    MaybeSentinel,
    Name,
    RemovalSentinel,
    RemoveFromParent,
)
from libcst import matchers as m
from libcst.codemod import CodemodContext, ContextAwareTransformer
from libcst.codemod.visitors import AddImportsVisitor
from libcst.metadata import ParentNodeProvider, Scope, ScopeProvider


class BaseDjCodemodTransformer(ContextAwareTransformer, ABC):
    deprecated_in: Tuple[int, int]
    removed_in: Tuple[int, int]


def module_matcher(
    import_parts: Sequence[str],
) -> Union[m.Attribute, m.Name]:
    """Build matcher for a module given sequence of import parts."""
    # If only one element, it is just a Name
    if len(import_parts) == 1:
        return m.Name(import_parts[0])
    *values, attr = import_parts
    value = module_matcher(values)
    return m.Attribute(value=value, attr=m.Name(attr))


def import_from_matches(node: ImportFrom, module_parts: Sequence[str]) -> bool:
    """Check if an `ImportFrom` node matches sequence of module parts."""
    return m.matches(node, m.ImportFrom(module=module_matcher(module_parts)))


class BaseRenameTransformer(BaseDjCodemodTransformer, ABC):
    """Base class to help rename or move a declaration."""

    rename_from: str
    rename_to: str

    simple_rename = True

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        *self.old_parent_module_parts, self.old_parent_name, _ = (
            *self.old_module_parts,
            self.old_name,
        ) = self.old_all_parts = self.rename_from.split(".")
        *self.new_parent_module_parts, self.new_parent_name, _ = (
            *self.new_module_parts,
            self.new_name,
        ) = self.new_all_parts = self.rename_to.split(".")
        self.ctx_key_import_scope = f"{self.rename_from}-import_scope"
        self.ctx_key_name_matcher = f"{self.rename_from}-name_matcher"
        self.ctx_key_new_func = f"{self.rename_from}-new_func"

    @property
    def name_matcher(self):
        return self.context.scratch.get(self.ctx_key_name_matcher, None)

    @property
    def new_func(self):
        return self.context.scratch.get(self.ctx_key_new_func, None)

    def leave_ImportFrom(
        self, original_node: ImportFrom, updated_node: ImportFrom
    ) -> Union[BaseSmallStatement, RemovalSentinel]:
        """Update import statements for matching old module name."""
        return (
            self._check_import_from_exact(original_node, updated_node)
            or self._check_import_from_parent(original_node, updated_node)
            or self._check_import_from_child(updated_node)
            or updated_node
        )

    def _check_import_from_exact(
        self, original_node: ImportFrom, updated_node: ImportFrom
    ) -> Optional[Union[BaseSmallStatement, RemovalSentinel]]:
        """
        Check for when the thing to replace is imported exactly.

        When `parent.module.the_thing` is transformed, detect such import:

            from parent.module import the_thing
        """
        # First, exit early if 'import *' is used
        if isinstance(updated_node.names, ImportStar):
            return None
        # Check whether the exact symbol is imported
        if not import_from_matches(updated_node, self.old_module_parts):
            return None
        # Match, update the node an return it
        new_import_aliases = []
        for import_alias in updated_node.names:
            if not self.old_name or import_alias.evaluated_name == self.old_name:
                if import_alias.evaluated_alias is None:
                    self.save_import_scope(original_node)
                    self.context.scratch[self.ctx_key_name_matcher] = m.Name(
                        self.old_name
                    )
                    if self.new_name:
                        self.context.scratch[self.ctx_key_new_func] = Name(
                            self.new_name
                        )
                if self.rename_from != self.rename_to:
                    if self.simple_rename:
                        AddImportsVisitor.add_needed_import(
                            context=self.context,
                            module=".".join(self.new_module_parts),
                            obj=self.new_name or import_alias.evaluated_name,
                            asname=import_alias.evaluated_alias,
                        )
                    continue
            new_import_aliases.append(import_alias)
        if not new_import_aliases:
            # Nothing left in the import statement: remove it
            return RemoveFromParent()
        # Some imports are left, update the statement
        new_import_aliases = clean_new_import_aliases(new_import_aliases)
        return updated_node.with_changes(names=new_import_aliases)

    def _check_import_from_parent(
        self, original_node: ImportFrom, updated_node: ImportFrom
    ) -> Optional[Union[BaseSmallStatement, RemovalSentinel]]:
        """
        Check for when the parent module of thing to replace is imported.

        When `parent.module.the_thing` is transformed, detect such import:

            from parent import module
        """
        # First, exit early if 'import *' is used
        if isinstance(updated_node.names, ImportStar):
            return None
        # Check whether parent module is imported
        if not import_from_matches(updated_node, self.old_parent_module_parts):
            return None
        # Match, update the node an return it
        new_import_aliases = []
        for import_alias in updated_node.names:
            if import_alias.evaluated_name == self.old_parent_name:
                self.save_import_scope(original_node)
                module_name_str = (
                    import_alias.evaluated_alias or import_alias.evaluated_name
                )
                self.context.scratch[self.ctx_key_name_matcher] = m.Attribute(
                    value=m.Name(module_name_str),
                    attr=m.Name(self.old_name),
                )
                self.context.scratch[self.ctx_key_new_func] = Attribute(
                    attr=Name(self.new_name),
                    value=Name(import_alias.evaluated_alias or self.new_parent_name),
                )
                if self.old_parent_module_parts != self.new_parent_module_parts:
                    # import statement needs updating
                    AddImportsVisitor.add_needed_import(
                        context=self.context,
                        module=".".join(self.new_parent_module_parts),
                        obj=self.new_parent_name,
                        asname=import_alias.evaluated_alias,
                    )
                    continue
            new_import_aliases.append(import_alias)
        if not new_import_aliases:
            # Nothing left in the import statement: remove it
            return RemoveFromParent()
        # Some imports are left, update the statement
        new_import_aliases = clean_new_import_aliases(new_import_aliases)
        return updated_node.with_changes(names=new_import_aliases)

    def _check_import_from_child(
        self, updated_node: ImportFrom
    ) -> Optional[Union[BaseSmallStatement, RemovalSentinel]]:
        """
        Check import of a member of the module being codemodded.

        When `parent.module.the_thing` is transformed, detect such import:

            from parent.module.thing import something
        """
        # First, exit early if 'import *' is used
        if isinstance(updated_node.names, ImportStar):
            return None
        # Check whether a member of the module is imported
        if not import_from_matches(updated_node, self.old_all_parts):
            return None
        # Match, add import for all imported names and remove the existing import
        for import_alias in updated_node.names:
            AddImportsVisitor.add_needed_import(
                context=self.context,
                module=".".join(self.new_all_parts),
                obj=import_alias.evaluated_name,
                asname=import_alias.evaluated_alias,
            )
        return RemoveFromParent()

    def resolve_parent_node(self, node: CSTNode) -> CSTNode:
        parent_nodes = self.context.wrapper.resolve(ParentNodeProvider)  # type: ignore
        return parent_nodes[node]

    def resolve_scope(self, node: CSTNode) -> Scope:
        scopes_map = self.context.wrapper.resolve(ScopeProvider)  # type: ignore
        return scopes_map[node]  # type: ignore

    def save_import_scope(self, import_from: ImportFrom) -> None:
        scope = self.resolve_scope(import_from)
        self.context.scratch[self.ctx_key_import_scope] = scope

    @property
    def import_scope(self) -> Optional[Scope]:
        return self.context.scratch.get(self.ctx_key_import_scope, None)

    def leave_Name(self, original_node: Name, updated_node: Name) -> BaseExpression:
        """Rename reference to the imported name."""
        matcher = self.name_matcher
        if (
            matcher
            and m.matches(updated_node, matcher)
            and not self.is_wrapped_in_call(original_node)
            and self.matches_import_scope(original_node)
        ):
            return updated_node.with_changes(value=self.new_name)
        return super().leave_Name(original_node, updated_node)

    def leave_Attribute(
        self, original_node: Attribute, updated_node: Attribute
    ) -> BaseExpression:
        matcher = self.name_matcher
        if (
            matcher
            and m.matches(updated_node, matcher)
            and not self.is_wrapped_in_call(original_node)
            and self.matches_import_scope(original_node)
        ):
            return updated_node.with_changes(
                value=Name(self.new_parent_name),
                attr=Name(self.new_name),
            )
        return super().leave_Attribute(original_node, updated_node)

    def is_wrapped_in_call(self, node: CSTNode) -> bool:
        """Check whether given node is wrapped in Call."""
        parent = self.resolve_parent_node(node)
        return m.matches(parent, m.Call())

    def matches_import_scope(self, node: CSTNode) -> bool:
        """Check whether given node matches the scope of the import."""
        try:
            scope = self.resolve_scope(node)
        except KeyError:
            # Can't resolve scope of node -> consider no match
            # Might be because of one of these reasons:
            # - It's the same name in another scope
            # - It's a attribute with the same name
            # - It's a keyword argument
            return False
        return scope == self.import_scope


def clean_new_import_aliases(
    import_aliases: List[ImportAlias],
) -> List[ImportAlias]:
    """Clean up a list of import aliases."""
    # Sort them
    cleaned_import_aliases = sorted(import_aliases, key=lambda n: n.evaluated_name)
    # Remove any trailing commas
    last_name = cleaned_import_aliases[-1]
    if last_name.comma != MaybeSentinel.DEFAULT:
        cleaned_import_aliases[-1] = last_name.with_changes(comma=MaybeSentinel.DEFAULT)
    return cleaned_import_aliases


class BaseFuncRenameTransformer(BaseRenameTransformer, ABC):
    """Base class to help rename or move a function."""

    def leave_Call(self, original_node: Call, updated_node: Call) -> BaseExpression:
        matcher = self.name_matcher
        if m.matches(updated_node, m.Call(func=matcher)):
            return self.update_call(updated_node=updated_node)
        return super().leave_Call(original_node, updated_node)

    def update_call(self, updated_node: Call) -> BaseExpression:
        updated_args = self.update_call_args(updated_node)
        return updated_node.with_changes(args=updated_args, func=self.new_func)

    def update_call_args(self, node: Call) -> Sequence[Arg]:
        return node.args



"""Some utilities to deal with files.

Inspired or taken from black:
https://github.com/psf/black
"""
from pathlib import Path
from typing import Iterable, Iterator, List, Optional, Set

from pathspec import PathSpec


def get_sources(src: Iterable[str]) -> List[Path]:
    """Return a list of sources to codemod.

    Based on a list of paths or files, recursively find python source files,
    ignoring the ones according to `.gitignore`.
    """
    root = find_project_root(src)
    gitignore = get_gitignore(root)
    sources: Set[Path] = set()
    for s in src:
        p = Path(s)
        paths: Iterable[Path] = []
        if p.is_file():
            paths = [p]
        elif p.is_dir():
            paths = list(p.iterdir())
        if paths:
            sources.update(gen_python_files(paths, root, gitignore))
    return sorted(sources)


def find_project_root(sources: Iterable[str]) -> Path:
    """Return a directory containing .git, .hg, or pyproject.toml.

    That directory will be a common parent of all files and directories
    passed in `sources`.

    If no directory in the tree contains a marker that would specify it's the
    project root, the root of the file system is returned.
    """
    if not sources:
        return Path("/").resolve()

    path_srcs = [Path(Path.cwd(), src).resolve() for src in sources]

    # A list of lists of parents for each 'src'. 'src' is included as a
    # "parent" of itself if it is a directory
    src_parents = [
        list(path.parents) + ([path] if path.is_dir() else []) for path in path_srcs
    ]

    common_base = max(
        set.intersection(*(set(parents) for parents in src_parents)),
        key=lambda path: path.parts,
    )

    for directory in (common_base, *common_base.parents):
        if (directory / ".git").exists():
            return directory

        if (directory / ".hg").is_dir():
            return directory

        if (directory / "pyproject.toml").is_file():
            return directory

    return directory


def get_gitignore(root: Path) -> PathSpec:
    """Return a PathSpec matching gitignore content if present."""
    gitignore = root / ".gitignore"
    lines: List[str] = []
    if gitignore.is_file():
        with gitignore.open() as gf:
            lines = gf.readlines()
    return PathSpec.from_lines("gitwildmatch", lines)


def gen_python_files(
    paths: Iterable[Path],
    root: Path,
    gitignore: PathSpec,
) -> Iterator[Path]:
    """Generate all files under `paths`.

    Files listed in .gitignore are not considered.
    """
    for child in paths:
        normalized_path = normalize_path(child, root)
        if normalized_path is None:
            continue

        # First ignore files matching .gitignore
        if gitignore.match_file(normalized_path):
            continue

        if child.is_dir():
            yield from gen_python_files(
                child.iterdir(),
                root,
                gitignore,
            )

        elif child.is_file() and str(child).endswith(".py"):
            yield child


def normalize_path(path: Path, root: Path) -> Optional[str]:
    """Normalize `path`. May return `None` if `path` was ignored."""
    try:
        abspath = path if path.is_absolute() else Path.cwd() / path
        normalized_path = abspath.resolve().relative_to(root).as_posix()
    except OSError as exc:
        print(f"{path} cannot be read because {exc}")
        return None
    except ValueError:
        if path.is_symlink():
            print(f"{path} is a symbolic link that points outside {root}")
            return None
        raise

    return normalized_path



from typing import Sequence, Tuple

from libcst import Arg, BaseExpression, Call, Name, SimpleString
from libcst import matchers as m
from libcst.codemod.visitors import AddImportsVisitor

from django_codemod.constants import DJANGO_3_0, DJANGO_4_0
from django_codemod.visitors.base import BaseFuncRenameTransformer


class PatternNotSupported(RuntimeError):
    pass


# A set of regex special characters sans the dash character
REGEX_SPECIALS_SANS_DASH = set("()[]{}?*+|^$\\.&~# \t\n\r\v\f")

REGEX_TO_CONVERTER = {
    "[0-9]+": "int",
    r"\d+": "int",
    ".+": "path",
    "[-a-zA-Z0-9_]+": "slug",
    r"[\w-]+": "slug",
    r"[-\w]+": "slug",
    "[^/]+": "str",
    "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}": "uuid",
}


class URLTransformer(BaseFuncRenameTransformer):
    """Update `django.conf.urls.url` by equivalent `path` or `re_path`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.conf.urls.url"
    rename_to = "django.urls.re_path"

    simple_rename = False

    def update_call(self, updated_node: Call) -> BaseExpression:
        """Update `url` call with either `path` or `re_path`."""
        try:
            return self.update_call_to_path(updated_node)
        except PatternNotSupported:
            # Safe fallback to re_path()
            AddImportsVisitor.add_needed_import(
                context=self.context,
                module=".".join(self.new_module_parts),
                obj=self.new_name,
            )
            return super().update_call(updated_node)

    def update_call_to_path(self, updated_node: Call) -> Call:
        """Update an URL pattern to `path()` in simple cases."""
        first_arg, *other_args = updated_node.args
        if not isinstance(first_arg.value, SimpleString):
            raise PatternNotSupported()
        # Extract the URL pattern from the first argument
        pattern = first_arg.value.evaluated_value
        # If we reach this point, we might be able to use `path()`
        call = self.build_path_call(pattern, other_args)
        AddImportsVisitor.add_needed_import(
            context=self.context,
            module=".".join(self.new_module_parts),
            obj="path",
        )
        return call

    def build_path_call(self, pattern: str, other_args: Sequence[Arg]) -> Call:
        """Build the `Call` node using Django 2.0's `path()` function."""
        route = self.build_route(pattern)
        updated_args = (Arg(value=SimpleString(f"'{route}'")), *other_args)
        return Call(args=updated_args, func=Name("path"))

    def build_route(self, pattern: str) -> str:
        """Build route from a URL pattern."""
        stripped_pattern = pattern.lstrip("^").rstrip("$")
        route = ""
        # Parse each group
        while "(?P<" in stripped_pattern:
            route_part, stripped_pattern = self.parse_next_group(stripped_pattern)
            route += route_part
        route += stripped_pattern
        self.check_route(route)
        return route

    @staticmethod
    def parse_next_group(left_to_parse: str) -> Tuple[str, str]:
        """Extract captured group info."""
        prefix, rest = left_to_parse.split("(?P<", 1)
        group, left_to_parse = rest.split(")", 1)
        group_name, group_regex = group.split(">", 1)
        try:
            converter = REGEX_TO_CONVERTER[group_regex]
        except KeyError:
            raise PatternNotSupported("No converter found")
        return prefix + f"<{converter}:{group_name}>", left_to_parse

    @staticmethod
    def check_route(route: str) -> None:
        """Check that route doesn't contain anymore regex."""
        if set(route) & REGEX_SPECIALS_SANS_DASH:
            raise PatternNotSupported(f"Route {route} contains regex")

    def update_call_args(self, node: Call) -> Sequence[Arg]:
        """Remove keyword argument from first argument of `re_path`."""
        first_arg, *other_args = node.args
        if m.matches(first_arg, m.Arg(keyword=m.Name("regex"))):
            first_arg = Arg(value=first_arg.value)
            return [first_arg, *other_args]
        return super().update_call_args(node)



from libcst import Attribute, BaseExpression, Call
from libcst import matchers as m

from django_codemod.constants import DJANGO_2_0, DJANGO_2_1, DJANGO_3_0, DJANGO_4_0
from django_codemod.visitors.base import (
    BaseDjCodemodTransformer,
    BaseFuncRenameTransformer,
)


class HttpUrlQuoteTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.http.urlquote` by `urllib.parse.quote`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.http.urlquote"
    rename_to = "urllib.parse.quote"


class HttpUrlQuotePlusTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.http.urlquote_plus` by `urllib.parse.quote_plus`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.http.urlquote_plus"
    rename_to = "urllib.parse.quote_plus"


class HttpUrlUnQuoteTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.http.urlunquote` by `urllib.parse.unquote`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.http.urlunquote"
    rename_to = "urllib.parse.unquote"


class HttpUrlUnQuotePlusTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.http.urlunquote_plus` by `urllib.parse.unquote_plus`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.http.urlunquote_plus"
    rename_to = "urllib.parse.unquote_plus"


class IsSafeUrlTransformer(BaseFuncRenameTransformer):
    """Rename `django.utils.http.is_safe_url` to `url_has_allowed_host_and_scheme`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.http.is_safe_url"
    rename_to = "django.utils.http.url_has_allowed_host_and_scheme"


class HttpRequestXReadLinesTransformer(BaseDjCodemodTransformer):
    """Replace `HttpRequest.xreadlines()` by iterating over the request."""

    deprecated_in = DJANGO_2_0
    removed_in = DJANGO_3_0

    # This should be conservative and only apply changes to:
    # - variables called `request`/`req`
    # - `request`/`req` attributes (e.g `self.request`/`view.req`...)
    matcher = m.Call(
        func=m.Attribute(
            value=m.OneOf(
                m.Name(value="request"),
                m.Name(value="req"),
                m.Attribute(attr=m.Name(value="request")),
                m.Attribute(attr=m.Name(value="req")),
            ),
            attr=m.Name(value="xreadlines"),
        )
    )

    def leave_Call(self, original_node: Call, updated_node: Call) -> BaseExpression:
        if not isinstance(updated_node.func, Attribute):
            # A bit redundant with matcher below,
            # but this is to make type checker happy
            return super().leave_Call(original_node, updated_node)
        if m.matches(updated_node, self.matcher):
            return updated_node.func.value
        return super().leave_Call(original_node, updated_node)


class CookieDateTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.http.cookie_date()` by `http_date`."""

    deprecated_in = DJANGO_2_1
    removed_in = DJANGO_3_0
    rename_from = "django.utils.http.cookie_date"
    rename_to = "django.utils.http.http_date"



from typing import List, Optional

from libcst import BaseExpression, Call, ImportFrom, ImportStar, MaybeSentinel, Module
from libcst import matchers as m

from django_codemod.constants import DJANGO_1_9, DJANGO_2_0
from django_codemod.visitors.base import BaseDjCodemodTransformer, import_from_matches


class SignalDisconnectWeakTransformer(BaseDjCodemodTransformer):
    """Remove the `weak` argument to `Signal.disconnect()`."""

    deprecated_in = DJANGO_1_9
    removed_in = DJANGO_2_0

    ctx_key_prefix = "SignalDisconnectWeakTransformer"
    ctx_key_call_matchers = f"{ctx_key_prefix}-call_matchers"
    builtin_signals = [
        "pre_init",
        "post_init",
        "pre_save",
        "post_save",
        "pre_delete",
        "post_delete",
        "m2m_changed",
        "pre_migrate",
        "post_migrate",
    ]
    import_alias_matcher = m.OneOf(
        *(m.ImportAlias(name=m.Name(signal_name)) for signal_name in builtin_signals)
    )

    @property
    def disconnect_call_matchers(self) -> List[m.Call]:
        return self.context.scratch.get(self.ctx_key_call_matchers, [])

    def add_disconnect_call_matcher(self, call_matcher: m.Call) -> None:
        self.context.scratch[
            self.ctx_key_call_matchers
        ] = self.disconnect_call_matchers + [call_matcher]

    def leave_Module(self, original_node: Module, updated_node: Module) -> Module:
        """Clear context when leaving module."""
        self.context.scratch.pop(self.ctx_key_call_matchers, None)
        return super().leave_Module(original_node, updated_node)

    def visit_ImportFrom(self, node: ImportFrom) -> Optional[bool]:
        """Set the `Call` matcher depending on which signals are imported.."""
        if not import_from_matches(
            node, ["django", "db", "models", "signals"]
        ) or isinstance(node.names, ImportStar):
            return False
        for import_alias in node.names:
            if m.matches(import_alias, self.import_alias_matcher):
                # We're visiting an import statement for a built-in signal
                # Get the actual name it's imported as (in case of import alias)
                imported_name_str = (
                    import_alias.evaluated_alias or import_alias.evaluated_name
                )
                # Add the call matcher for the current signal to the list
                self.add_disconnect_call_matcher(
                    m.Call(
                        func=m.Attribute(
                            value=m.Name(imported_name_str),
                            attr=m.Name("disconnect"),
                        ),
                    )
                )
        return None

    def leave_Call(self, original_node: Call, updated_node: Call) -> BaseExpression:
        """
        Remove the `weak` argument if present in the call.

        This is only changing calls with keyword arguments.
        """
        if self.disconnect_call_matchers and m.matches(
            updated_node, m.OneOf(*self.disconnect_call_matchers)
        ):
            updated_args = []
            should_change = False
            last_comma = MaybeSentinel.DEFAULT
            # Keep all arguments except the one with the keyword `weak` (if present)
            for index, arg in enumerate(updated_node.args):
                if m.matches(arg, m.Arg(keyword=m.Name("weak"))):
                    # An argument with the keyword `weak` was found
                    # -> we need to rewrite the statement
                    should_change = True
                else:
                    updated_args.append(arg)
                last_comma = arg.comma  # type: ignore
            if should_change:
                # Make sure the end of line is formatted as initially
                updated_args[-1] = updated_args[-1].with_changes(comma=last_comma)
                return updated_node.with_changes(args=updated_args)
        return super().leave_Call(original_node, updated_node)



from typing import Optional

from libcst import Arg
from libcst import matchers as m


def find_keyword_arg(args, keyword_name: str) -> Optional[Arg]:
    matcher = m.Arg(keyword=m.Name(keyword_name))
    for arg in args:
        if m.matches(arg, matcher):
            return arg
    return None



from typing import Generator, Optional, Sequence, Union

from libcst import (
    Assign,
    AssignTarget,
    Decorator,
    FlattenSentinel,
    ImportFrom,
    ImportStar,
    Module,
    Name,
    RemovalSentinel,
)
from libcst import matchers as m

from django_codemod.constants import DJANGO_1_9, DJANGO_2_0
from django_codemod.visitors.base import BaseDjCodemodTransformer, import_from_matches


class AssignmentTagTransformer(BaseDjCodemodTransformer):
    """Replace `assignment_tag` by `simple_tag`."""

    deprecated_in = DJANGO_1_9
    removed_in = DJANGO_2_0

    ctx_key_prefix = "AssignmentTagTransformer"
    ctx_key_library_call_matcher = f"{ctx_key_prefix}-library_call_matcher"
    ctx_key_decorator_matcher = f"{ctx_key_prefix}-decorator_matcher"

    @property
    def library_call_matcher(self) -> Optional[m.Call]:
        return self.context.scratch.get(self.ctx_key_library_call_matcher, None)

    @property
    def decorators_matcher(self) -> Optional[m.BaseMatcherNode]:
        return self.context.scratch.get(self.ctx_key_decorator_matcher, None)

    def leave_Module(self, original_node: Module, updated_node: Module) -> Module:
        """Clear context when leaving module."""
        self.context.scratch.pop(self.ctx_key_library_call_matcher, None)
        self.context.scratch.pop(self.ctx_key_decorator_matcher, None)
        return super().leave_Module(original_node, updated_node)

    def visit_ImportFrom(self, node: ImportFrom) -> Optional[bool]:
        """Record whether an interesting import is detected."""
        import_matcher = (
            # django.template
            self._template_import_matcher(node)
            # django.template.Library
            or self._library_import_matcher(node)
        )
        if import_matcher:
            self.context.scratch[self.ctx_key_library_call_matcher] = import_matcher
        return None

    def _template_import_matcher(self, node: ImportFrom) -> Optional[m.Call]:
        """Return matcher if django.template is imported."""
        imported_name_str = self._get_imported_name(node, "django.template")
        if not imported_name_str:
            return None
        # Build the `Call` matcher to look out for, e.g. `template.Library()`
        return m.Call(
            func=m.Attribute(
                attr=m.Name("Library"), value=m.Name(value=imported_name_str)
            )
        )

    def _library_import_matcher(self, node: ImportFrom) -> Optional[m.Call]:
        """Return matcher if django.template.Library is imported."""
        imported_name_str = self._get_imported_name(node, "django.template.Library")
        if not imported_name_str:
            return None
        # Build the `Call` matcher to look out for, e.g. `Library()`
        return m.Call(func=m.Name(imported_name_str))

    @staticmethod
    def _get_imported_name(node: ImportFrom, import_path: str) -> Optional[str]:
        """Resolve the imported name if present."""
        if isinstance(node.names, ImportStar):
            return None
        *modules, name = import_path.split(".")
        if not import_from_matches(node, modules):
            return None
        for import_alias in node.names:
            if m.matches(import_alias, m.ImportAlias(name=m.Name(name))):
                # We're visiting the import statement we're looking for
                # Get the actual name it's imported as (in case of import alias)
                imported_name_str = (
                    import_alias.evaluated_alias or import_alias.evaluated_name
                )
                return imported_name_str
        return None

    def visit_Assign(self, node: Assign) -> Optional[bool]:
        """Record variable name the `Library()` call is assigned to."""
        if self.library_call_matcher and m.matches(
            node,
            m.Assign(value=self.library_call_matcher),
        ):
            # Visiting a `register = template.Library()` statement
            # Generate decorator matchers based on left hand side names
            decorator_matchers = self._gen_decorator_matchers(node.targets)
            # should match if any of the decorator matches
            self.context.scratch[self.ctx_key_decorator_matcher] = m.OneOf(
                *decorator_matchers
            )
        return super().visit_Assign(node)

    @staticmethod
    def _gen_decorator_matchers(
        assign_targets: Sequence[AssignTarget],
    ) -> Generator[m.Decorator, None, None]:
        """Generate matchers for all possible decorators."""
        for assign_target in assign_targets:
            # for each variable it's assigned to
            if isinstance(assign_target.target, Name):
                # get the name of the target
                target_str = assign_target.target.value
                # matcher we should use for finding decorators to modify
                yield m.Decorator(
                    decorator=m.Attribute(
                        value=m.Name(target_str),
                        attr=m.Name("assignment_tag"),
                    )
                )

    def leave_Decorator(
        self, original_node: Decorator, updated_node: Decorator
    ) -> Union[Decorator, FlattenSentinel[Decorator], RemovalSentinel]:
        """Update decorator call if all conditions are met."""
        if self.decorators_matcher and m.matches(updated_node, self.decorators_matcher):
            # If we have a decorator matcher, and it matches,
            # then update the node with new name
            updated_decorator = updated_node.decorator.with_changes(
                attr=Name("simple_tag")
            )
            return updated_node.with_changes(decorator=updated_decorator)
        return super().leave_Decorator(original_node, updated_node)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_codemod', 'django_codemod.utils', 'django_codemod.visitors']

package_data = \
{'': ['*']}

install_requires = \
['click<8', 'libcst<=0.3.18', 'pathspec>=0.6,<1', 'rich<11']

extras_require = \
{'docs': ['Sphinx>=3.3.0,<4.0.0',
          'sphinx-rtd-theme>=0.5.0,<0.6.0',
          'myst-parser>=0.13.0,<0.14.0']}

entry_points = \
{'console_scripts': ['djcodemod = django_codemod.cli:djcodemod']}

setup_kwargs = {
    'name': 'django-codemod',
    'version': '1.4.1',
    'description': 'A command line tool to automatically fix Django deprecations.',
    'long_description': '<p align="center">\n  <img alt="Django Codemod" src="https://raw.githubusercontent.com/browniebroke/django-codemod/main/assets/djangocodemod-stroke.png">\n</p>\n\n<p align="center">\n  <a href="https://github.com/browniebroke/django-codemod/actions?query=workflow%3ACI">\n    <img alt="CI Status" src="https://img.shields.io/github/workflow/status/browniebroke/django-codemod/CI?label=CI&logo=github&style=flat-square">\n  </a>\n  <a href="https://django-codemod.readthedocs.io">\n    <img src="https://img.shields.io/readthedocs/django-codemod.svg?logo=read-the-docs&logoColor=fff&style=flat-square" alt="Documentation Status">\n  </a>\n  <a href="https://codecov.io/gh/browniebroke/django-codemod">\n    <img src="https://img.shields.io/codecov/c/github/browniebroke/django-codemod.svg?logo=codecov&logoColor=fff&style=flat-square" alt="Test coverage percentage">\n  </a>\n</p>\n<p align="center">\n  <a href="https://python-poetry.org/">\n    <img src="https://img.shields.io/badge/packaging-poetry-299bd7?style=flat-square&logo=data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAA4AAAASCAYAAABrXO8xAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAJJSURBVHgBfZLPa1NBEMe/s7tNXoxW1KJQKaUHkXhQvHgW6UHQQ09CBS/6V3hKc/AP8CqCrUcpmop3Cx48eDB4yEECjVQrlZb80CRN8t6OM/teagVxYZi38+Yz853dJbzoMV3MM8cJUcLMSUKIE8AzQ2PieZzFxEJOHMOgMQQ+dUgSAckNXhapU/NMhDSWLs1B24A8sO1xrN4NECkcAC9ASkiIJc6k5TRiUDPhnyMMdhKc+Zx19l6SgyeW76BEONY9exVQMzKExGKwwPsCzza7KGSSWRWEQhyEaDXp6ZHEr416ygbiKYOd7TEWvvcQIeusHYMJGhTwF9y7sGnSwaWyFAiyoxzqW0PM/RjghPxF2pWReAowTEXnDh0xgcLs8l2YQmOrj3N7ByiqEoH0cARs4u78WgAVkoEDIDoOi3AkcLOHU60RIg5wC4ZuTC7FaHKQm8Hq1fQuSOBvX/sodmNJSB5geaF5CPIkUeecdMxieoRO5jz9bheL6/tXjrwCyX/UYBUcjCaWHljx1xiX6z9xEjkYAzbGVnB8pvLmyXm9ep+W8CmsSHQQY77Zx1zboxAV0w7ybMhQmfqdmmw3nEp1I0Z+FGO6M8LZdoyZnuzzBdjISicKRnpxzI9fPb+0oYXsNdyi+d3h9bm9MWYHFtPeIZfLwzmFDKy1ai3p+PDls1Llz4yyFpferxjnyjJDSEy9CaCx5m2cJPerq6Xm34eTrZt3PqxYO1XOwDYZrFlH1fWnpU38Y9HRze3lj0vOujZcXKuuXm3jP+s3KbZVra7y2EAAAAAASUVORK5CYII=" alt="Poetry">\n  </a>\n  <a href="https://github.com/ambv/black">\n    <img src="https://img.shields.io/badge/code%20style-black-000000.svg?amp;style=flat-square" alt="black">\n  </a>\n  <a href="https://github.com/pre-commit/pre-commit">\n    <img src="https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white&style=flat-square" alt="pre-commit">\n  </a>\n</p>\n<p align="center">\n  <a href="https://pypi.org/project/django-codemod/">\n    <img src="https://img.shields.io/pypi/v/django-codemod.svg?logo=python&logoColor=fff&style=flat-square" alt="PyPi Status">\n  </a>\n  <img src="https://img.shields.io/pypi/pyversions/django-codemod.svg?style=flat-square&logo=python&amp;logoColor=fff" alt="pyversions">\n  <img src="https://img.shields.io/pypi/l/django-codemod.svg?style=flat-square" alt="license">\n  <a href="https://github.com/browniebroke/django-codemod">\n    <img src="https://tokei.rs/b1/github/browniebroke/django-codemod/" alt="LoC">\n  </a>\n</p>\n\nA tool to help upgrade Django projects to newer version of the framework by automatically fixing deprecations.\n\n## The problem\n\nWhen maintaining a Django site, over time you\'ll find yourself to a point where you\'ll need to update to the next major version of Django. When Django APIs changes, functions move or are removed, changing usages in your project might add up to many changes. Often these changes are simple to do, but sometimes a simple "find and replace" is not possible.\n\nTake, for instance, the removal of the `url()` function from Django 4.0, to be replaced by `re_path()`. In simple cases, you might even want to switch to `path()`, which has a nicer API. A typical Django project easily has 100\'s or routes, so this simple decision becomes a much longer task when to be made for each of them.\n\n### This solution\n\nThis project solves this problem by providing codemodders for simple changes like this. A codemodder re-writes your code from the old way to the new way.\n\nWith the help of AST analysis, we\'re able to understand what modifications are applicable, remove imports as they become irrelevant, and add missing ones as they are needed.\n\nTo continue the example, the tool will look at the route in the `url()` call, and decide whether the regular expression may be replaced by one of the built-in URL converters and use `path()` or stick to a regex and use `re_path()`.\n\nInterested? Check out [the documentation](https://django-codemod.readthedocs.io) for usage and the full list of codemodders.\n\n### What this tool is not\n\n- This tool is best suited for Django sites, NOT for reusable Django applications. The project needs to target a single Django version, e.g. 3.1.x.\n- You do NOT need to install this tool as part of your project dependencies, it is a CLI tool, not a Django package to be installed in your site.\n\n## Contributors ✨\n\nThanks goes to these wonderful people ([emoji key](https://allcontributors.org/docs/en/emoji-key)):\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n<table>\n  <tr>\n    <td align="center"><a href="https://browniebroke.com"><img src="https://avatars1.githubusercontent.com/u/861044?v=4?s=80" width="80px;" alt=""/><br /><sub><b>Bruno Alla</b></sub></a><br /><a href="https://github.com/browniebroke/django-codemod/commits?author=browniebroke" title="Code">💻</a> <a href="https://github.com/browniebroke/django-codemod/commits?author=browniebroke" title="Documentation">📖</a> <a href="#ideas-browniebroke" title="Ideas, Planning, & Feedback">🤔</a></td>\n    <td align="center"><a href="https://akx.github.io/"><img src="https://avatars2.githubusercontent.com/u/58669?v=4?s=80" width="80px;" alt=""/><br /><sub><b>Aarni Koskela</b></sub></a><br /><a href="https://github.com/browniebroke/django-codemod/commits?author=akx" title="Code">💻</a> <a href="#ideas-akx" title="Ideas, Planning, & Feedback">🤔</a> <a href="https://github.com/browniebroke/django-codemod/commits?author=akx" title="Tests">⚠️</a></td>\n    <td align="center"><a href="https://adamj.eu/"><img src="https://avatars2.githubusercontent.com/u/857609?v=4?s=80" width="80px;" alt=""/><br /><sub><b>Adam Johnson</b></sub></a><br /><a href="https://github.com/browniebroke/django-codemod/commits?author=adamchainz" title="Documentation">📖</a></td>\n    <td align="center"><a href="https://sobolevn.me"><img src="https://avatars1.githubusercontent.com/u/4660275?v=4?s=80" width="80px;" alt=""/><br /><sub><b>Nikita Sobolev</b></sub></a><br /><a href="https://github.com/browniebroke/django-codemod/commits?author=sobolevn" title="Documentation">📖</a></td>\n    <td align="center"><a href="http://www.zapier.com"><img src="https://avatars3.githubusercontent.com/u/21158438?v=4?s=80" width="80px;" alt=""/><br /><sub><b>Chris VanderKolk</b></sub></a><br /><a href="https://github.com/browniebroke/django-codemod/commits?author=cvanderkolk" title="Code">💻</a></td>\n    <td align="center"><a href="https://ghuser.io/jayvdb"><img src="https://avatars1.githubusercontent.com/u/15092?v=4?s=80" width="80px;" alt=""/><br /><sub><b>John Vandenberg</b></sub></a><br /><a href="https://github.com/browniebroke/django-codemod/issues?q=author%3Ajayvdb" title="Bug reports">🐛</a> <a href="https://github.com/browniebroke/django-codemod/commits?author=jayvdb" title="Code">💻</a></td>\n    <td align="center"><a href="https://iamshnoo.github.io/blog/"><img src="https://avatars1.githubusercontent.com/u/45921510?v=4?s=80" width="80px;" alt=""/><br /><sub><b>Anjishnu</b></sub></a><br /><a href="#infra-iamshnoo" title="Infrastructure (Hosting, Build-Tools, etc)">🚇</a></td>\n  </tr>\n  <tr>\n    <td align="center"><a href="https://github.com/drewbrew"><img src="https://avatars1.githubusercontent.com/u/7773256?v=4?s=80" width="80px;" alt=""/><br /><sub><b>Drew Winstel</b></sub></a><br /><a href="#ideas-drewbrew" title="Ideas, Planning, & Feedback">🤔</a></td>\n  </tr>\n</table>\n\n<!-- markdownlint-restore -->\n<!-- prettier-ignore-end -->\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\nThis project follows the [all-contributors](https://github.com/all-contributors/all-contributors) specification. Contributions of any kind welcome!\n\n## Credits\n\nThis package was created with\n[Cookiecutter](https://github.com/audreyr/cookiecutter) and the\n[audreyr/cookiecutter-pypackage](https://github.com/audreyr/cookiecutter-pypackage)\nproject template.\n',
    'author': 'Bruno Alla',
    'author_email': 'alla.brunoo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/browniebroke/django-codemod',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
